// This is a simple SOAP server.

#include "inkey.ch"
#include "web.ch"
#INCLUDE "dcdialog.CH"
#INCLUDE "fileio.CH"
#INCLUDE "error.CH"

#define CRLF Chr(13)+Chr(10)

#Pragma Library("dcxml.lib")

#define PORT 8080

#define ERR_LOG_FILE      'ERROR.LOG'

#xtranslate NTrim(<n>)   =>  LTrim(Str(<n>))


STATIC soTLog

PROCEDURE Main()

LOCAL cPort, nKey, oHttpEndpoint

cPort := AllTrim(Str(PORT))
? "SOAP Server sample."
?
? "With SOAPUI navigate to:"
?
? "     http://localhost:"+cPort+"/SOAP/wsdl"
?
? "Press ESC to quit"

//
// The listener thread is established
//
oHttpEndpoint := MyHttpEndpoint():new( PORT )
oHttpEndpoint:start()

soTLog := DC_TLog():new("SoapServer.Log",GetLogFields())
soTLog:software := 'Soap Server (Xbase++)'
soTLog:version := '1.00'
soTLog:start()
nKey := 0
DO WHILE nKey <> K_ESC
   nKey := Inkey(1)
ENDDO
soTLog:stop()

//
// Stop the endpoint before the application
// is terminated
//
oHttpEndpoint:stop()

RETURN

* ==============

CLASS MyHttpEndpoint FROM HttpEndpoint

EXPORTED:

VAR tlog
METHOD defaultProcessors
METHOD start

ENDCLASS

METHOD MyHttpEndpoint:start()

RETURN SUPER:start()

* ---------

METHOD MyHttpEndpoint:defaultProcessors()

LOCAL cProcessors := { ;
   "MyWebHandlerProcessor", ;
   "MyFileProcessor" }

RETURN cProcessors

* ------------

CLASS MyWebHandlerProcessor FROM WebHandlerProcessor

EXPORTED:

CLASS METHOD Run()

ENDCLASS

* ---------

CLASS METHOD MyWebHandlerProcessor:Run( cFileName, oWebAppHost, oCxpApplication, ;
                                        oHttpRequest, oHttpResponse )

LOCAL xReturn, oError, nSeconds := Seconds(), nTimeTaken

GetLogData(oHttpRequest)
SoapAction('')

ErrorBlock({|o|Break(o)})

BEGIN SEQUENCE

xReturn := SUPER:run( cFileName, oWebAppHost, oCxpApplication, ;
                      oHttpRequest, oHttpResponse )

nTimeTaken := Round((Seconds()-nSeconds)*1000,0)

RECOVER USING oError

  oHttpResponse:contentType := 'text/html'
  oHttpResponse:writeStr( oError:cargo )
  oHttpResponse:send()

END SEQUENCE

soTLog:fieldData[6] := SoapAction()
soTLog:fieldData[7] := 200
soTLog:fieldData[11] := nTimeTaken

soTLog:write()

RETURN xReturn

* ------------

CLASS MyFileProcessor FROM HtmlFileProcessor

EXPORTED:

CLASS METHOD MimeTypes()
CLASS METHOD FileTypes()
CLASS METHOD Run()

ENDCLASS

* ---------

CLASS METHOD MyFileProcessor:Run( cFileName, oWebAppHost, oCxpApplication, ;
                                  oHttpRequest, oHttpResponse )

LOCAL x, oError, nTimeTaken

GetLogData(oHttpRequest)

BEGIN SEQUENCE

x := SUPER:run( cFileName, oWebAppHost, oCxpApplication, ;
                oHttpRequest, oHttpResponse )

RECOVER USING oError

 oHttpResponse:contentType := 'text/html'
 oHttpResponse:writeStr( oError:cargo )
 oHttpResponse:send()

END SEQUENCE

soTLog:fieldData[7] := 200
soTLog:fieldData[11] := nTimeTaken

soTLog:write()

RETURN x

* -----------

CLASS METHOD MyFileProcessor:MimeTypes()

LOCAL cMimeTypes := { ;
   "text/html", ;
   "text/htm", ;
   "text/text", ;
   "text/xml", ;
   "text/wsdl", ;
   "image/jpg", ;
   "image/gif", ;
   "image/png", ;
   "application/pdf"}

RETURN cMimeTypes

* -----------

CLASS METHOD MyFileProcessor:FileTypes()

LOCAL cFileTypes := { ;
   ".html", ;
   ".htm", ;
   ".txt", ;
   ".xml", ;
   ".wsdl", ;
   ".jpg", ;
   ".gif", ;
   ".png", ;
   ".pdf"}

RETURN cFileTypes

* ==============

CLASS Soap FROM WebHandler
  EXPORTED:
    VAR envelope
    VAR rootNode
    VAR headerNode
    VAR bodyNode
    METHOD GetDateTime
    METHOD UploadXML
    METHOD Services
    METHOD execute
    METHOD errorResponse
    METHOD response
    METHOD wsdl
    METHOD getHttpVar
    METHOD getNode
    METHOD getVar
    METHOD getFormVars
    METHOD getFormVar
    METHOD Process

ENDCLASS

* ---------

METHOD Soap:execute( cMethodName )

LOCAL cResponse := "", oError

BEGIN SEQUENCE

IF IsMethod(SELF, cMethodName)
  cResponse := SUPER:execute( cMethodName )
ELSE
  cResponse := ::errorResponse( cMethodName )
ENDIF

RECOVER USING oError

END SEQUENCE
IF Empty(oError)
  RETURN cResponse
ENDIF

RETURN LogError(oError)

* ----------

METHOD Soap:errorResponse( cMethodName, cMessage )

LOCAL cResponse
DEFAULT cMessage := 'No handler for: ' + cMethodName

PRIVATE cErrorMessage := cMessage
PRIVATE cErrorMethod := cMethodName

TEXT INTO cResponse
<?xml version="1.0" ?>
<env:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:env="http://schemas.xmlsoap.org/soap/envelope/" xmlns:enc="http://schemas.xmlsoap.org/soap/encoding/" xmlns:xs="http://www.w3.org/2001/XMLSchema" env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<env:Body>
<m:FaultResponse xmlns:m="http://donnay-software.com/soap">
  &cErrorMessage
</m:FaultResponse>
</env:Body>
</env:Envelope>
ENDTEXT

RETURN cResponse

* ----------

METHOD Soap:process( cMethodName )

LOCAL cXml, cResponse := '<resp>Thank you</resp>'

cXml := ::httpRequest:getContent()

cResponse := myfunction( cXml )

RETURN cResponse

* ----------

METHOD Soap:response( cMethodName, cMessage )

LOCAL cResponse

DEFAULT cMessage := 'Message Received'

PRIVATE cMethodResponse := cMethodName + 'Response'
PRIVATE cMessageResponse := cMessage

TEXT INTO cResponse WRAP
<?xml version="1.0" ?>
<env:Envelope
     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
     xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
     xmlns:enc="http://schemas.xmlsoap.org/soap/encoding/"
     xmlns:xs="http://www.w3.org/2001/XMLSchema"
     env:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
<env:Body>
<m:&cMethodResponse xmlns:m="http://donnay-software.com/soap">
  &cMessageResponse
</m:&cMethodResponse>
</env:Body>
</env:Envelope>
ENDTEXT

RETURN cResponse

* ----------

METHOD Soap:services()

LOCAL cResponse, cContent, cSoapAction

cContent := ::httpRequest:content

::envelope := cContent

::rootNode := DC_Xml2ObjectTree(cContent)
IF ::rootNode == NIL
  RETURN ::errorResponse('Fault','Invalid SOAP Envelope')
ENDIF

cSoapAction := ::getHttpVar('SOAPAction')

IF Empty(cSoapAction)
  ::headerNode := ::rootNode:findNode('Header',,.t.)
  ::bodyNode := ::rootNode:findNode('Body',,.t.)
  IF Valtype(::bodyNode) == 'O'
    cSoapAction := ::bodyNode:childList[1]:name
    IF ':' $ cSoapAction
      cSoapAction := Substr(cSoapAction,At(':',cSoapAction)+1)
    ENDIF
  ENDIF
ENDIF

IF !Empty(cSoapAction)
  cSoapAction := Strtran(cSoapAction,'"','')
  SoapAction(cSoapAction)
  IF !IsMethod(self,cSoapAction)
    RETURN ::errorResponse('Fault','No SOAP method: ' + cSoapAction)
  ELSE
    cResponse := self:&(cSoapAction)(::rootNode:childList[1])
  ENDIF
ELSE
  cResponse := ::errorResponse('<!---UNKNOWN--->')
ENDIF

RETURN cResponse

* -----------

METHOD Soap:wsdl()

RETURN MemoRead('SoapServices.wsdl')

* -----------

METHOD Soap:getDateTime()

RETURN ::response('GetDateTime', Dtos(Date()) + ' ' + Time())

* -----------

METHOD Soap:UploadXML( oMainNode )

LOCAL oNode, cResponse, cString, cMethod := 'UploadXML'

oNode := oMainNode:findNode('UserName',,.t.)
IF oNode == NIL
  RETURN ::errorResponse( cMethod,'No "UserName" Tag in Soap Request')
ENDIF
IF oNode:content # 'Alaska'
  RETURN ::errorResponse( cMethod,'Invalid UserName')
ENDIF

oNode := oMainNode:findNode('Password',,.t.)
IF oNode == NIL
  RETURN ::errorResponse( cMethod,'No "Password" Tag in Soap Request')
ENDIF
IF oNode:content # 'PolarBear'
  RETURN ::errorResponse( cMethod,'Invalid Password')
ENDIF

oNode := oMainNode:findNode('XmlDocument',,.t.)

IF oNode == NIL
  RETURN ::errorResponse( cMethod,'No "XMLDocument" Tag in Soap Request')
ENDIF

cString := oNode:asString()

MemoWrit( Dtos(Date())+Strtran(Time(),':','')+'.Xml',cString )

RETURN ::response( cMethod, 'XML Successfully received' )

* -----------

METHOD Soap:getHttpVar( cName )

LOCAL cMangleName, cValue

cMangleName := StrTran( cName, "-", "_" )
cValue := ::HttpRequest:getVariable( cMangleName )
IF cValue == NIL
  RETURN ""
ENDIF

RETURN cValue

* ---------

METHOD Soap:getVar( cTagName, aAttr )

LOCAL oNode

oNode := ::getNode(cTagName)
IF Valtype(oNode) == 'O'
  IF Empty(oNode:childList())
    RETURN oNode:contents
  ELSE
    RETURN oNode:childList()
  ENDIF
ENDIF

RETURN oNode

* ---------

METHOD Soap:getNode( cTagName )

LOCAL oNode

IF Valtype(::headerNode) == 'O'
  oNode := ::headerNode:findNode(cTagName,,.t.)
  IF Valtype(oNode) == 'O'
    RETURN oNode
  ENDIF
ENDIF

IF Valtype(::bodyNode) == 'O'
  oNode := ::bodyNode:findNode(cTagName,,.t.)
  IF Valtype(oNode) == 'O'
    RETURN oNode
  ENDIF
ENDIF

RETURN oNode

* ----------

METHOD Soap:getFormVars()

LOCAL aVars := ::httpRequest:getForm():getCollection()

RETURN aVars

* ----------

METHOD Soap:getFormVar( cVar )

LOCAL nFound, aVars := ::httpRequest:getForm():getCollection()

nFound := AScan( aVars, {|a|Upper(a[1])==Upper(cVar)})
IF nFound > 0
  RETURN aVars[nFound,2]
ENDIF

RETURN nil

* ----------

FUNCTION MyFunction( cXml )

LOCAL oNode, cSenderName, cUserName, oRootNode := DC_Xml2ObjectTree(cXml), ;
      cSoapAction

oNode := oRootNode:findNode('SenderNameCode',,.t.)
IF Valtype(oNode) == 'O'
  cSenderName := oNode:content
ELSE
  cSenderName := 'Error'
ENDIF

oNode := oRootNode:findNode('userName',,.t.)
IF Valtype(oNode) == 'O'
  cUserName := oNode:content
ELSE
  cUserName := 'Error'
ENDIF

RETURN '<response>Hello ' + cUserName + ' : ' + cSenderName + '</response>'

* ========================

CLASS DC_TLog

EXPORTED:

VAR software
VAR fieldNames
VAR fieldData
VAR logFileName
VAR whiteSpace
VAR date
VAR version
VAR isActive
VAR fileHandle

METHOD Init, Start, Stop, Write, WriteHeader, DateTime

ENDCLASS

METHOD DC_TLog:init( cLogFile, aFieldNames )

::fieldNames := aFieldNames
::fieldData := Array(Len(aFieldNames))
::logFileName := cLogFile
::whiteSpace := ' '
::isActive := .f.
::fileHandle := 0

RETURN self

METHOD DC_TLog:start()

LOCAL nHandle

IF !FExists(::logFileName)
  nHandle := FCreate(::logFileName)
  FClose(nHandle)
ENDIF
nHandle := FOpen(::logFileName,FO_WRITE+FO_SHARED)
FSeek(nHandle,0,FS_END)

IF nHandle <= 0
  RETURN .f.
ENDIF
::fileHandle := nHandle

RETURN ::writeHeader()

METHOD DC_TLog:writeHeader()

LOCAL i

::date := Date()

IF !Empty(::software) .AND. Valtype(::software) == 'C'
  FWrite(::fileHandle,'#Software: ' + ::software + CRLF)
ENDIF
IF !Empty(::version) .AND. Valtype(::version) == 'C'
  FWrite(::fileHandle,'#Version: ' + ::version + CRLF)
ENDIF
FWrite(::fileHandle,'#Date: ' + ::DateTime() + CRLF)
IF Valtype(::fieldNames) == 'A'
  Fwrite(::fileHandle,'#Fields: ')
  FOR i := 1 TO Len(::fieldNames)
    Fwrite(::fileHandle,::fieldNames[i]+::whiteSpace)
  NEXT
  Fwrite(::fileHandle,CRLF)
ENDIF

::isActive := ::fileHandle > 0

RETURN ::isActive

METHOD DC_TLog:write( cFields )

LOCAL i

IF Valtype(cFields) == 'C'
  FWrite(::fileHandle,cFields+CRLF)
ELSE
  FOR i := 1 TO Len(::fieldData)
    FWrite(::fileHandle,XtoC(::fieldData[i])+::whiteSpace)
  NEXT
  FWrite(::fileHandle,CRLF)
ENDIF

RETURN ::fileHandle > 0

METHOD DC_TLog:stop()

FClose(::fileHandle)

::fileHandle := 0
::isActive := .f.

RETURN .t.

METHOD DC_TLog:DateTime()

RETURN Alltrim(Str(Day(Date()))) + '-' + Left(CMonth(Date()),3) + '-' + Str(Year(Date()),4) + ' ' + Time()

* -----------

GETSETFUNCTION SoapAction DEFAULT ''

* -----------

STATIC FUNCTION GetLogData(oHttpRequest)

soTLog:fieldData[1] := Time()
soTLog:fieldData[2] := oHttpRequest:GetVariable('REMOTE_ADDR')
soTLog:fieldData[3] := oHttpRequest:GetVariable('REQUEST_METHOD')
soTLog:fieldData[4] := oHttpRequest:GetVariable('REQUEST_URI')
soTLog:fieldData[5] := oHttpRequest:GetVariable('QUERY_STRING')
soTLog:fieldData[6] := '' // sc-soapaction (see below)
soTLog:fieldData[7] := 0 // sc-status (see below)
soTLog:fieldData[8] := 0 // sc-win32-status
soTLog:fieldData[9] := 0 // sc-bytes
soTLog:fieldData[10] := oHttpRequest:GetContentLength()
soTLog:fieldData[11] := 0 // time-taken
soTLog:fieldData[12] := oHttpRequest:GetVariable('SERVER_PROTOCOL')
soTLog:fieldData[13] := oHttpRequest:GetVariable('HTTP_USER_AGENT')

RETURN nil

* ------------

STATIC FUNCTION GetLogFields()

LOCAL aFieldNames := ;
   { 'time', ;             // 1
     'c-ip', ;             // 2
     'cs-method', ;        // 3
     'cs-uri-stem', ;      // 4
     'cs-uri-query', ;     // 5
     'cs-soapaction', ;    // 6
     'sc-status', ;        // 7
     'sc-win32-status', ;  // 8
     'sc-bytes', ;         // 9
     'cs-bytes', ;         // 10
     'time-taken', ;       // 11
     'cs-version', ;       // 12
     'cs(Referer)' }       // 13

RETURN aFieldNames

* ------------

STATIC FUNCTION LogError( oError )

LOCAL cError, i, cProcName

cError := "==== Donnay Software SOAP Server Error ====" + CRLF +;
             "Date: " + DTOS(Date()) + ", Time: " + Time() + CRLF +;
             "App: " + AppName(.t.) + ", Thread: " + LTrim(Str(ThreadID())) + CRLF

// error generated while processing http server request
cError += if(oError:severity > ES_WARNING, "ERROR ", "WARNING " ) +;
          if(ValType(oError:subSystem)   == "C", oError:subSystem, "???" ) + "/" +;
          if(ValType(oError:subCode)     == "N", NTrim(oError:subCode), "???" ) + CRLF +;
          if(ValType(oError:cargo)       == "C", oError:cargo + CRLF, "" ) +;
          if(ValType(oError:description) == "C", oError:description, "" )

// add either filename or operation
if !Empty(oError:filename)
   cError += ": " + oError:filename
elseif !Empty(oError:operation)
   cError += ": " + oError:operation
endif

if !Empty( oError:osCode )
   cError += CRLF + "(OS Error: " + NTrim(oError:osCode) + ")"
endif

for i := 2 to 12
   if Empty(cProcName := ProcName(i))
      exit
   endif
   cError += CRLF+"Called from "+cProcName+"("+LTrim(Str(ProcLine(i)))+")"
next

SaveToFile( cError + CRLF, ERR_LOG_FILE )

cError := '<Fault>' + CRLF + cError + '</Fault>' + CRLF

RETURN cError

* -----------

STATIC FUNCTION SaveToFile( cText, cFileName )

Local i := 0, nHandle

if ValType(cText) != "C" .or. Empty(cText)
   Return .f.
endif

if FExists(cFileName)
   while (nHandle := FOpen(cFileName, FO_WRITE + FO_DENYWRITE)) < 0
      if ++i > 9
         Return .f.
      endif
      Sleep(20)
   end
   FSeek(nHandle, 0, FS_END)  // position to end of file
else
   while (nHandle := FCreate(cFileName)) < 0
      if ++i > 9
         Return .f.
      endif
      Sleep(20)
   end
endif

FWrite(nHandle, cText + CRLF)
FClose(nHandle)

Return .t.


