/*
 ķ
   Program..: _DCTXT.PRG                                               
   Author...: Roger J. Donnay                                          
   Notice...: (c) DONNAY Software Designs 1987-1998                    
   Date.....: Sep  2, 1998                                             
   Notes....: Text file functions                                      
                                                                       
   Functions:  dc_txtopen(), dc_txtclose(), dc_txtword(), dc_txteof(), 
               dc_txtskip(), dc_txtline(), dc_txtgoto(), dc_txtcount(),
               dc_txtlno(), dc_txtsize(), dc_txtfind()                 
 Ľ
*/

#INCLUDE "dcdialog.ch"
#INCLUDE "fileio.ch"

STATIC aText := {}

FUNCTION dc_txtopen ( cFileName, nMode )

LOCAL nHandle := FOpen(cFileName, nMode)

IF nHandle > 0
  IF Len(aText) < nHandle
    ASize(aText,nHandle)
  ENDIF
  aText[nHandle] := 1
ENDIF

RETURN nHandle

* -----------------

FUNCTION dc_txtclose ( nHandle )

nHandle := IIF(Valtype(nHandle)='N',nHandle,0)
IF nHandle <= 0
  RETURN .f.
ENDIF
IF Len(aText) < nHandle
  ASize(aText,nHandle)
ENDIF
aText[nHandle] := 0
RETURN FCLOSE( nHandle )

* -----------------

FUNCTION dc_txtword ( cString, nPosition, cWhiteSpace )

LOCAL nStart, nEnd, nCount, i, cChar

cString := IIF(Valtype(cString)='C',cString,'')
nPosition := IIF(Valtype(nPosition)='N',nPosition,0)
cWhiteSpace := IIF(Valtype(cWhiteSpace)='C',cWhiteSpace,' ')

nCount := 0

i := 1
nStart := 0
nEnd := 0
DO WHILE i <= LEN(cString)
  cChar := Substr(cString,i,1)
  IF nStart=0 .AND. !(cChar$cWhiteSpace)
    nCount++
    nStart := i
    nEnd := 0
  ELSEIF nStart>0 .AND. (cChar$cWhiteSpace .OR. i=LEN(cString))
    IF i = LEN(cString)
      nEnd := i
    ELSE
      nEnd := i-1
    ENDIF
    IF nCount = nPosition
      RETURN SubStr(cString,nStart,nEnd-nStart+1)
    ENDIF
    nStart := 0
  ENDIF
  i++
ENDDO
RETURN ''


/*
IF !(SUBSTR(cString,1,1)==cWhiteSpace)
  cString := cWhiteSpace + cString
ENDIF
nStart := DC_AtNext( cWhiteSpace, cString, nPosition )
nEnd := DC_AtNext( cWhiteSpace, cString, nPosition+1 )
RETURN SubStr( cString, nStart+1, nEnd-nStart )
*/

* -----------------

FUNCTION dc_txteof ( nHandle )

LOCAL nEnd,nCurrent

DEFAULT nHandle := 0
IF nHandle <= 0
  RETURN .f.
ENDIF
nCurrent := FSeek(nHandle,0,1)
nEnd := FSeek(nHandle,0,2) - 1
FSeek(nHandle,nCurrent,0)
RETURN nCurrent>=nEnd

* -----------------

FUNCTION dc_txtskip ( nHandle, nLines, cEol, nLineLen )

LOCAL cString, cChar, nPos, nEnd, i

cEol := IIF( Valtype(cEol)='C',cEol,Chr(13)+Chr(10))

DEFAULT nHandle := 0, ;
        nLines := 1, ;
        nLineLen := 40000

IF nHandle <= 0
  RETURN 0
ENDIF
IF Len(aText)<nHandle
  ASize(aText,nHandle)
  aText[nHandle] := 0
ENDIF
FOR i := 1 TO nLines
  nPos := FSeek( nHandle, 0, 1 ) // Save current pointer
  cString := Space(nLineLen)
  FRead( nHandle, @cString, nLineLen )
  nEnd := At( cEol, cString ) // End of line
  IF nEnd=0
    nPos := FSeek( nHandle, 0, 2)
    EXIT
  ENDIF
  FSeek( nHandle, nPos, 0 ) // Go back to saved position
  nPos := FSeek( nHandle, nPos+nEnd, 0 ) // Go to start of next line
NEXT
aText[nHandle] += (i-1)
RETURN aText[nHandle]

* -----------------

FUNCTION dc_txtline ( nHandle, cEol, nLineLen )

LOCAL cString, cChar, nPos, nEnd, i

DEFAULT nLineLen := 40000

cEol := IIF( Valtype(cEol)='C',cEol,Chr(13)+Chr(10))

DEFAULT nHandle := 0
IF nHandle <= 0 .OR. Len(aText) < nHandle .OR. Valtype(aText[nHandle])#'N'
  RETURN ''
ENDIF

nPos := FSeek( nHandle, 0, 1 )
cString := Space(nLineLen)
FRead( nHandle, @cString, nLineLen )

nEnd := At( cEol, cString )
FSeek( nHandle, nPos, 0 )
IF nEnd=0
  RETURN cString
ENDIF
cString := SubStr( cString, 1, nEnd-1)
FOR i := 1 TO Len(cEol)
  cString := StrTran( cString,SubStr(cEol,i,1),'')
NEXT
RETURN cString

* -----------------

FUNCTION dc_txtgoto ( nHandle, nLineNumber )

LOCAL i

DEFAULT nLineNumber := 0, nHandle := 0

IF nHandle <= 0 .OR. Len(aText) < nHandle .OR. Valtype(aText[nHandle])#'N'
  RETURN .f.
ENDIF
FSEEK( nHandle, 0, 0 )
aText[nHandle] := 1
FOR i := 1 TO nLineNumber - 1
  DC_TxtSkip( nHandle, 1 )
  IF DC_TxtEof(nHandle)
    RETURN .f.
  ENDIF
NEXT
RETURN .t.

* -----------------

FUNCTION dc_txtlno ( nHandle )

DEFAULT nHandle := 0
IF nHandle <= 0 .OR. Len(aText) < nHandle .OR. Valtype(aText[nHandle])#'N'
  RETURN 0
ENDIF
RETURN aText[nHandle]

* -----------------

FUNCTION dc_txtcount ( nHandle, cEol )

LOCAL nPos, nEnd, nCount, cString, nCurrent

cEol := IIF( Valtype(cEol)='C',cEol,Chr(13)+Chr(10))

nHandle := IIF(Valtype(nHandle)='N',nHandle,0)
IF nHandle <= 0
  RETURN 0
ENDIF
nCurrent := FSeek(nHandle,0,1)
FSeek(nHandle,0,0)
nCount := 0
DO WHILE !DC_TxtEof(nHandle)
  nPos := FSeek( nHandle, 0, 1 )
  cString := Space(1000)
  FRead( nHandle, @cString, 1000 )
  nEnd := At( Left(cEol,1), cString )
  nCount++
  IF nEnd=0
    EXIT
  ENDIF
  FSeek( nHandle, nPos, 0 )
  nPos := FSeek( nHandle, nPos+nEnd, 0 )
ENDDO
FSeek(nHandle,nCurrent,0)
RETURN  nCount

* -----------------

FUNCTION dc_txtsize ( nHandle )

LOCAL nCurrent, nEnd
nCurrent := FSeek(nHandle,0,1)
nEnd := FSeek(nHandle,0,2)
FSeek(nHandle,nCurrent,0)
RETURN nEnd

* -----------------

FUNCTION dc_txtfind( nHandle, cSearchStr, lCase )

LOCAL nMemBlock := 16384
LOCAL cBuffer := SPACE(nMemBlock), ;
      nBytesRead := nMemBlock, nFoundAt := 0, lFound := .f.,;
      nOffSet := 0

lCase := IIF( Valtype(lCase)='L',lCase,.t. )

DO WHILE nBytesRead=nMemBlock .AND. !lFound
  nOffSet    := FSEEK(nHandle,0,1)
  nBytesRead := FREAD(nHandle,@cBuffer,nMemBlock)
  nFoundAt   := AT( IIF(!lCase,UPPER(cSearchStr),cSearchstr),;
                IIF(!lCase,UPPER(cBuffer),cBuffer))
  lFound     := nFoundAt # 0
ENDDO
nFoundAt := IIF(lFound, nOffSet + nFoundAt,-1)
IF lFound
  FSeek( nHandle, nFoundAt, 0 )
ENDIF
RETURN nFoundAt

* -------------------

FUNCTION DC_FileRead( cFileName )

LOCAL nTop, nBottom, nHandle, cString

nHandle := Fopen(cFileName)
IF nHandle <= 0
   MsgBox('File ' + cFileName + ' cannot be opened')
   RETURN ''
ENDIF

nBottom := FSeek( nHandle, 0 , FS_END )
nTop := FSeek( nHandle, 0 , FS_SET )

cString := Space(nBottom-nTop-1)
FRead( nHandle, @cString, nBottom - nTop - 1 )

FClose(nHandle)

RETURN cString
